/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file   lcmaps_afs.c
 *  \brief  Interface to the LCMAPS plugins
 *  \author Gerben Venekamp for the EU DataGrid.
 *
 *  This file contains the code for an example LCMAPS plugin and shows
 *  the interface the plugin has to provide to the LCMAPS. The
 *  interface consists of the following functions:
 *  -# plugin_initialize()
 *  -# plugin_run()
 *  -# plugin_terminate()
 *  -# plugin_introspect()
 *
 */


/*!
 *  \page lcmaps_afs.mod afs plugin
 *
 *  \section afssection1 SYNOPSIS
 *  \b  lcmaps_afs.mod [ -lifetime \<hh[:mm[:ss]]\> | -server \<server\> | -port \<port\> | -cell \<cell\> | -setpag | -gsiklog ]
 *
 *  \section afssection2 DESCRIPTION
 *
 *  This plugin is an Acquisition and Enforcement Plugin and provides the LCMAPS with
 *  afs access. Users must have an accessible afs file syteme somewhere.
 *  The plugin can only function when gssklog and gssklogd are installed.
 *  Underneath it uses the client program gssklog to get an afs token using the user's proxy.
 *  The gssklog client depends on the daemon gssklogd to get the afs
 *  token. The daemon must be installed one of the afs servers and hence
 *  there is no necessity for it to be installed on the same machine
 *  as the gatekeeper or the client side of gssklog. The afs machine must
 *  be reachable from the client machine.
 *  The gssklogd daemon uses the file /etc/grid-security/afsgrid-mapfile for the DN-to-AFSaccount
 *  mapping.
 *
 *  \section afsoptions OPTIONS
 *  \subsection afsoptie1 -lifetime \<ticket lifetime in hh[:mm[:ss]]\>
 *    Use this option to specify the maximum lifetime of afs token.
 *    gssklog allows a maximum of 720 hours. Defaults to 12 hours.
 *
 *  \subsection afsoptie2 -server \<afs server\>
 *  Name of the server where gssklogd runs.  
 *
 *  \subsection afsoptie3 -port \<port\>
 *  Port number at which the gssklod daemon runs. Defaults to 750.
 *
 *  \subsection afsoptie4 -cell \<cell\>
 *  Cell name.
 *
 *  \subsection afsoption5 -setpag
 *  Setpag.
 *
 *  \subsection afsoption6 -gsiklog
 *  Use GSI interface instead of GSS interface
 *
 *  \section afsReturnvalue RETURN VALUES
 *  \li LCMAPS_MOD_SUCCESS : Success
 *  \li LCMAPS_MOD_FAIL    : Failure
 *
 *  \section afsErrors ERRORS
 *  See bugzilla for known errors (http://marianne.in2p3.fr/datagrid/bugzilla/)
 *
 *  \section afsSeeAlso SEE ALSO
 *  \ref lcmaps_localaccount.mod "lcmaps_localaccount.mod",
 *  \ref lcmaps_poolaccount.mod "lcmaps_poolaccount.mod",
 *  \ref lcmaps_posix_enf.mod "lcmaps_posix_enf.mod",
 *  \ref lcmaps_ldap_enf.mod "lcmaps_ldap_enf.mod",
 *  \ref lcmaps_voms.mod "lcmaps_voms.mod",
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <gssapi.h>

#include "lcmaps_afs_config.h"

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>

/******************************************************************************
                                Definitions
******************************************************************************/
#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1

/*  Default values for options.  */
static const char* default_lifetime = "12:00";

/*  The following option may be set.  */
static char* lifetime = NULL;
static char* port     = NULL;
static char* server   = NULL;
static char* gsiklog  = NULL;
static char* cell     = NULL;
static char* setpag   = NULL;

static int argv_count = 0;

/******************************************************************************
                          Module specific prototypes
******************************************************************************/
static int plugin_run_or_verify(int, lcmaps_argument_t *, int);

int run_cmd(char* cmd, char* argv[]);
int fail_afs(void);
int check_argument(const char* name, char** value, unsigned int num, int argc, char* argv[], int* pos);
int construct_argv(char* cmd, char** argv[]);
void destruct_argv(char** argv[]);


/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(int *argc, lcmaps_argument_t **argv)
{
  static lcmaps_argument_t argList[] = {
    { "job_request", "lcmaps_request_t",  1, NULL},
    { "user_cred"  , "gss_cred_id_t"   ,  0, NULL},
    { NULL         , NULL              , -1, NULL}
  };
  
  lcmaps_log_debug(5, "lcmaps_afs: plugin_introspect(): introspecting\n");
  
  *argv = argList;
  *argc = lcmaps_cntArgs(argList);
  lcmaps_log_debug(5, "lcmaps_afs: plugin_introspect(): address first argument: 0x%x\n",argList);
  
  return LCMAPS_MOD_SUCCESS;
}

/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
    Everything that is needed to initialize the plugin should be put
    inside this function.  Arguments as read from the LCMAPS database
    (argc, argv) are passed to the plugin.
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(int argc, char **argv)
{
  int i;
  
  lcmaps_log_debug(5, "lcmaps_afs: plugin_initialize(): passed arguments:\n");
  for (i=1; i<argc; i++) {
    if (argv[i][0] != '-') {
      lcmaps_log(LOG_NOTICE, "lcmaps_afs: plugin_initialize(): Could not interpret arg[%d] = '%s'.\n", i, argv[i]);
      lcmaps_log(LOG_INFO,   "lcmaps_afs:                    : Skipping argument...\n");
      continue;
    }

    lcmaps_log_debug(5, "lcmaps_afs: plugin_initialize(): arg %d is %s\n", i, argv[i]);

    if (check_argument("-lifetime", &lifetime, 1, argc, argv, &i) == LCMAPS_MOD_FAIL) return LCMAPS_MOD_FAIL; else continue;
    if (check_argument("-server",   &server,   1, argc, argv, &i) == LCMAPS_MOD_FAIL) return LCMAPS_MOD_FAIL; else continue;
    if (check_argument("-cell",     &cell,     1, argc, argv, &i) == LCMAPS_MOD_FAIL) return LCMAPS_MOD_FAIL; else continue;
    if (check_argument("-port",     &port,     1, argc, argv, &i) == LCMAPS_MOD_FAIL) return LCMAPS_MOD_FAIL; else continue;
    if (check_argument("-setpag",   &setpag,   0, argc, argv, &i) == LCMAPS_MOD_FAIL) return LCMAPS_MOD_FAIL; else continue;
    if (check_argument("-gsiklog",  &gsiklog,  0, argc, argv, &i) == LCMAPS_MOD_FAIL) return LCMAPS_MOD_FAIL; else continue;
  }

  return LCMAPS_MOD_SUCCESS;
}


/*!
 *  check_argument(name, value, num, argc, argv, pos)
 *
 *  Check if the name matches the argument at position pos.
 *  If it does, then depending if num is equal to zero or one,
 *  zero or one values are read and stored at value.
 *
 *  \param name   Name to be matched.
 *  \param value  store the value, may not be NULL!
 *  \param num    Number of args to be read, following a match.
 *  \param argc   Argc variable.
 *  \param argv   Argv variable.
 *  \param pos    Current possition in the argv list.
 *
 *  \return 1 if successfull, 0 otherwise.
 *
 */
int check_argument(const char* name, char** value, unsigned int num,
		   int argc, char* argv[], int* pos)
{
  /*
   *  Value should be a NULL pointer, if not, it should point to
   *  allocated memory; in which case memory is freeed and the
   *  pointer set to NULL.
   */
  if (*value) {
    free(*value);
    *value = NULL;
  }

  /*  Check if the argument is recognized.  */
  if (strcmp(argv[*pos], name)==0) {
    lcmaps_log(LOG_ERR, "lcmaps_afs: plugin_initialize(): Found argument %s.\n", name);
    /*  Does the argument need a value?  */
    if (num==1) {
      /*  Protect ourselves to reading past the argv list.  */
      if ((*pos+num)<(argc)) {
	if (!(*value = strdup(argv[*pos+1]))) {
	  lcmaps_log(LOG_ERR, "lcmaps_afs: plugin_initialize(): Could not allocate memory.\n");
	  return LCMAPS_MOD_FAIL;
	}
	lcmaps_log(LOG_ERR, "lcmaps_afs:                    : With value = '%s'\n", *value);
      } else {
	/*
	 *  We have to have a value for the given argument. Since,
	 *  we have reached the end of the argv list the sysadmin
	 *  must have forgotten to specify it. We could silently
	 *  ignor this mistake and supply a default value. However
	 *  it is best that the sysadmin corrects this problem.
	 */
	lcmaps_log(LOG_ERR, "lcmaps_afs: plugin_initialize(): Could not read additional values.\n");
	lcmaps_log(LOG_ERR, "lcmaps_afs:                    : Please correct configuration file.\n");
	return LCMAPS_MOD_FAIL;
      }
    } else if (!(*value = strdup(name))) {
      lcmaps_log(LOG_ERR, "lcmaps_afs: plugin_initialize(): Could not allocate memory.\n");
      return LCMAPS_MOD_FAIL;
    }

    *pos += num+1;
    argv_count += num+1;
  }

  return LCMAPS_MOD_SUCCESS;
}


int plugin_run(int argc, lcmaps_argument_t *argv)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}

/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(
        int argc,
        lcmaps_argument_t * argv
)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(
        int argc,
        lcmaps_argument_t * argv,
        int lcmaps_mode
)
{
  char*  cmd      = "gssklog";
  char** cmd_argv = NULL;
  char*  logstr   = "lcmaps_afs: plugin_run()";

  gss_cred_id_t *pcred = NULL;
  gss_cred_id_t  cred  = GSS_C_NO_CREDENTIAL;

  if (lcmaps_mode == PLUGIN_RUN)
      logstr = "lcmaps_plugin_afs-plugin_run()";
  else if (lcmaps_mode == PLUGIN_VERIFY)
      logstr = "lcmaps_plugin_afs-plugin_verify()";
  else
  {
      lcmaps_log(LOG_ERR,
          "lcmaps_plugin_afs-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n",
          lcmaps_mode);
      return -1;
  }
  lcmaps_log_debug(4,"%s:\n", logstr);

  if (getuid()==0) {
    lcmaps_log(LOG_ERR, "%s: Afs plugin is run as root.\n", logstr);
    lcmaps_log(LOG_ERR, "%s: Exit afs plugin.\n", logstr);

    return -1;
  }

  /* Fetch user gss credential */
  if ((pcred = (gss_cred_id_t *)lcmaps_getArgValue("user_cred", "gss_cred_id_t", argc, argv))) {
    lcmaps_log_debug(5, "%s: Address user_cred: %p\n", logstr, pcred);
    cred = *pcred;

    if (cred == GSS_C_NO_CREDENTIAL) {
      lcmaps_log(LOG_ERR, "%s: User gss credential is empty.\n", logstr);
      lcmaps_log(LOG_ERR, "%s: Exit afs plugin.\n", logstr);

      return fail_afs();
    }
  } else {
    lcmaps_log(LOG_ERR, "%s: Could not get address of user_cred.\n", logstr);
    lcmaps_log(LOG_ERR, "%s: Exit afs plugin.\n", logstr);

    return fail_afs();
  }

  if (cred) {
    gss_buffer_desc  deleg_proxy_filename;
    OM_uint32        major_status = 0;
    OM_uint32        minor_status = 0;
        
    major_status = gss_export_cred(&minor_status,
				   cred,
				   NULL,
				   1,
				   &deleg_proxy_filename);

    if (major_status == GSS_S_COMPLETE) {
      char *cp;

      cp = strchr((char *)deleg_proxy_filename.value, '=');
      *cp = '\0';
      cp++;
      setenv((char *)deleg_proxy_filename.value, cp, 1);
      free(deleg_proxy_filename.value);
    } else {
      char *error_str = NULL;
      globus_object_t *error_obj;
      
      error_obj = globus_error_get((globus_result_t)minor_status);
      
      error_str = globus_error_print_chain(error_obj);
      
      lcmaps_log(LOG_ERR, "%s: Error, gss_export_cred(): %s\n", logstr, error_str);
    }
  }

  /* for gssapi_ssleay if we received delegated proxy certificate
   * they will be in a file in tmp pointed at by the 
   * X509_USER_DELEG_PROXY env variable. 
   * This will be owned by the current uid, we need this owned by
   * the user.
   * for other gss, this is a noop, since X509_USER_DELEG_PROXY 
   * should not be defined.
   */

  /*
   * the proxyfile is probably already owned by the current user,
   * so commented out
   */
//    {
//        char *proxyfile;
//        if ((proxyfile = getenv("X509_USER_PROXY")) != NULL)
//        {
//            chown(proxyfile,uid[0],priGid[0]);
//        }
//    }
  
  lcmaps_log_debug(5, "lcmaps_afs: plugin_run(): address first argument: 0x%x\n", argv);

  if (construct_argv(cmd, &cmd_argv) && (run_cmd(cmd, cmd_argv) != 0)) {
  lcmaps_log(LOG_ERR, "lcmaps_afs: construct_argv: cmd_argv = %x.\n", cmd_argv);
    destruct_argv(&cmd_argv);
    return fail_afs();
  }

  destruct_argv(&cmd_argv);

  //  rc = system(cmd);
  
  //  switch(rc) {
  //  case 0:   /*  Everything seems to be okay.  */
  //    break;
  //  default:
  //    lcmaps_log(LOG_ERR, "%s: gssklog returned unknown error: rc = %d", logstr, rc);
  //  }

  /*  succes  */
  if(getenv("X509_USER_PROXY")) {
    remove(getenv("X509_USER_PROXY"));
  }

  return LCMAPS_MOD_SUCCESS;
}


/*!
 *  construct_arfv(cmd, argv)
 *
 *  Construct an argv list. The first element is the name of the
 *  executable.
 *
 *  \param cmd  command to be executed, this will be argv[0].
 *  \param argv pointer that hold the location of argv[] pointer.
 *
 *  \returns 1 if successfull, 0 otherwise.
 *
 */
int construct_argv(char* cmd, char** argv[])
{
  int i;

  lcmaps_log_debug(5, "lcmaps_afs: construct_argv: Max number of elements of argv = %d.\n", argv_count);

  if (!(*argv = (char**)calloc(argv_count+2, sizeof(char *)))) {
    lcmaps_log(LOG_ERR, "lcmaps_afs: construct_argv: Could not allocate memory.\n");
    return 0;
  }

  i = 0;
  if (!((*argv)[i++] = strdup(cmd))) {
    lcmaps_log(LOG_ERR, "lcmaps_afs: construct_argv: Could not allocate memory.\n");
    return 0;
  }

  if (!((*argv)[i++] = strdup("-lifetime"))) {
    lcmaps_log(LOG_ERR, "lcmaps_afs: construct_argv: Could not allocate memory.\n");
    return 0;
  }

  if (lifetime)
    (*argv)[i++] = lifetime;
  else {
    if (!((*argv)[i++] = strdup(default_lifetime))) {
      lcmaps_log(LOG_ERR, "lcmaps_afs: construct_argv: Could not allocate memory.\n");
      return 0;
    }
  }

  lcmaps_log_debug(5, "lcmaps_afs: construct_argv: Constructed argv:\n");
  i = 0;
  while ((*argv)[i]) {
    lcmaps_log_debug(5, "lcmaps_afs: construct_argv: arg[%d] = '%s'.\n",i, (*argv)[i]);
    i++;
  }

  return 1;
}


/*!
 *  destruct_argv(argv)
 *
 *  Destruct the argv list by freeing all elements as well as the
 *  list itself.
 *
 *  \param argv argv list to be freed.
 *
 */
void destruct_argv(char** argv[])
{
  int i;

  if (*argv) {
    for (i=0; i<argv_count; i++) {
      free(argv[i]);
    }
    free(argv);

    *argv = NULL;
  }
}


/*!
 *  run_cmd(cmd, argv)
 *
 *  Run a command. The standard output and standard error are piped
 *  back into the current process and appended to the logging of
 *  the current process.
 *
 */
int run_cmd(char* cmd, char* argv[])
{
  int i, rc = 0;
  int child_status;
  pid_t term;
  char c;
  FILE* fr;
  pid_t pid;
  int pipe_fd[2];
  char *logstr = "lcmaps_afs: run_cmd()";

  struct sigaction sa, old_sa;

  sa.sa_handler = SIG_DFL;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;

  if (sigaction(SIGCHLD, &sa, &old_sa) == -1) {
    lcmaps_log(LOG_ERR, "%s: sigaction failed: %s\n", logstr, strerror(errno));
    return 1;
  }

  //  Create the read and write descriptors for the pipe.
  if (pipe(pipe_fd) < 0 ) {
    lcmaps_log(LOG_ERR, "%s: pipe failed: %s\n", logstr, strerror(errno));
    return 1;
  }

  lcmaps_log(LOG_ERR, "%s: executing: '%s'\n", logstr, cmd);
  for (i=1; argv[i]!=NULL; ++i)
    lcmaps_log(LOG_ERR, "%s:          :  argv[%d]: '%s'\n", logstr, i, argv[i]);
  
  lcmaps_log(LOG_ERR, "%s: -- begin output of '%s':\n", logstr, cmd);

  if ((pid = fork()) < 0) {
    //  Oops, it's an error!
    lcmaps_log(LOG_ERR, "%s: fork failed: %s\n", logstr, strerror(errno));
    return -1;
  } if (pid == 0) {
    //  Child process.
    close(pipe_fd[0]);
    dup2(pipe_fd[1], 1);   //  Redirect stdin
    dup2(pipe_fd[1], 2);   //  Redirect stderr
    close(pipe_fd[1]);

    execvp(cmd, argv);  //  Does not return when successful.

    lcmaps_log(LOG_ERR, "%s: execution of %s failed.\n", logstr, cmd);
    lcmaps_log(LOG_ERR, "%s:    %s.\n", logstr, strerror(errno));

    exit(-1);
  }

  //  Parent process.
  close(pipe_fd[1]);

  fr = fdopen(pipe_fd[0], "r");

  lcmaps_log(LOG_ERR, "lcmaps_afs:   gssklog: ");
  c = fgetc(fr);
  if (c != EOF) {
    do {
      if (c == '\n') {
	fprintf(stderr, "%c", c);
	lcmaps_log(LOG_ERR, "lcmaps_afs:   gssklog: ");
      } else
	fprintf(stderr, "%c", c);
    } while ((c = fgetc(fr))!=EOF);
  }

  fclose(fr);

  fprintf(stderr, "\n");

  lcmaps_log(LOG_ERR, "%s: -- end output of '%s'\n", logstr, cmd);

  if ((term = wait(&child_status)) > 0) {
    if (WIFEXITED(child_status))
      lcmaps_log_debug(5, "%s: '%s' exit status: rc = %d\n", logstr, cmd, WEXITSTATUS(child_status));
    else if (WIFSIGNALED(child_status))
      lcmaps_log_debug(2, "%s: '%s' terminated by signal %d\n", logstr, cmd, WTERMSIG(child_status));
    else if (WIFSTOPPED(child_status))
      lcmaps_log_debug(2, "%s: '%s' stopped by signal %d\n", logstr, cmd, WSTOPSIG(child_status));
    else
      lcmaps_log(LOG_ERR, "%s: '%s' abnormal exit.\n", logstr, cmd);
  } else {
    lcmaps_log(LOG_ERR, "%s: waiting for '%s' failed:\n", logstr, cmd);
    lcmaps_log(LOG_ERR, "%s: reason: %s\n", logstr, strerror(errno));
    rc = 1;
  }

  if (sigaction(SIGCHLD, &old_sa, NULL) == -1) {
    lcmaps_log(LOG_ERR, "%s: sigaction failed: %s\n", logstr, strerror(errno));
    lcmaps_log(LOG_ERR, "%s:                 : Old sigaction value has not been restored.\n", logstr);
    return 1;
  }

  close(pipe_fd[0]);

  return 0;
}


/*!
 *  fail_afs(void)
 *
 *  Do some necessary cleanup when the afs plugin fails.
 *
 */
int fail_afs(void) {

  if(getenv("X509_USER_PROXY")) {
    remove(getenv("X509_USER_PROXY"));
  }
  
  return LCMAPS_MOD_FAIL;
}


int plugin_terminate()
{
  lcmaps_log_debug(5,"lcmaps_afs: plugin_terminate(): terminating\n");
  
  return LCMAPS_MOD_SUCCESS;
}
